import datetime
import os
import shutil
from pairoptions import pairoptions
from EnumDefinitions import *
from EnumToString import EnumToString
from ismodelselfconsistent import ismodelselfconsistent
from marshall import marshall
from waitonlock import waitonlock
from loadresultsfromcluster import loadresultsfromcluster
import MatlabFuncs as m

def solve(md,solutionstring,*args):
	"""
	SOLVE - apply solution sequence for this model
 
	   Usage:
	      md=solve(md,solutionstring,varargin)
	      where varargin is a list of paired arguments of string OR enums
 
		solution types available comprise:
		 - 'StressbalanceSolution'		or 'sb'
		 - 'MasstransportSolution'		or 'mt'
		 - 'ThermalSolution'				or 'th'
		 - 'SteadystateSolution'		or 'st'
		 - 'TransientSolution'			or 'tr'
		 - 'BalancethicknessSolution' or 'mc'
		 - 'BalancevelocitySolution'	or 'bv'
		 - 'BedSlopeSolution'			or 'bs'
		 - 'SurfaceSlopeSolution'		or 'ss'
		 - 'HydrologySolution'			or 'hy'
		 - 'DamageEvolutionSolution'	or 'da'

 
	   extra options:
	      - loadonly : does not solve. only load results
		  - checkconsistency : 'yes' or 'no' (default is 'yes'), ensures checks on consistency of model
		  - restart: 'directory name (relative to the execution directory) where the restart file is located.
 
	   Examples:
	      md=solve(md,'StressbalanceSolution');
			md=solve(md,'sb');
	"""

	#recover and process solve options
	if solutionstring.lower() == 'sb' or solutionstring.lower() == 'stressbalance' or solutionstring.lower() == 'stressbalancesolution':
		solutionstring = 'StressbalanceSolution';
	elif solutionstring.lower() == 'mt' or solutionstring.lower() == 'masstransport' or solutionstring.lower() == 'masstransportsolution':
		solutionstring = 'MasstransportSolution';	
	elif solutionstring.lower() == 'th' or solutionstring.lower() == 'thermal' or solutionstring.lower() == 'thermalsolution':
		solutionstring = 'ThermalSolution';
	elif solutionstring.lower() == 'st' or solutionstring.lower() == 'steadystate' or solutionstring.lower() == 'steadystatesolution':
		solutionstring = 'SteadystateSolution';
	elif solutionstring.lower() == 'tr' or solutionstring.lower() == 'transient' or solutionstring.lower() == 'transientsolution':
		solutionstring = 'TransientSolution';
	elif solutionstring.lower() == 'mc' or solutionstring.lower() == 'balancethickness' or solutionstring.lower() == 'balancethicknesssolution':
		solutionstring = 'BalancethicknessSolution';
	elif solutionstring.lower() == 'bv' or solutionstring.lower() == 'balancevelocity' or solutionstring.lower() == 'balancevelocitysolution':
		solutionstring = 'BalancevelocitySolution';
	elif solutionstring.lower() == 'bs' or solutionstring.lower() == 'bedslope' or solutionstring.lower() == 'bedslopesolution':
		solutionstring = 'BedSlopeSolution';
	elif solutionstring.lower() == 'ss' or solutionstring.lower() == 'surfaceslope' or solutionstring.lower() == 'surfaceslopesolution':
		solutionstring = 'SurfaceSlopeSolution';
	elif solutionstring.lower() == 'hy' or solutionstring.lower() == 'hydrology' or solutionstring.lower() == 'hydrologysolution':
		solutioneum = 'HydrologySolution';
	elif solutionstring.lower() == 'da' or solutionstring.lower() == 'damageevolution' or solutionstring.lower() == 'damageevolutionsolution':
		solutionstring = 'DamageEvolutionSolution';
	else: 	
		raise ValueError("solutionstring '%s' not supported!" % solutionstring)
	options=pairoptions('solutionstring',solutionstring,*args)

	#recover some fields
	md.private.solution=solutionstring
	cluster=md.cluster 
	if m.strcmpi(options.getfieldvalue('batch','no'),'yes'):
		batch=1
	else:
		batch=0;

	#check model consistency
	if m.strcmpi(options.getfieldvalue('checkconsistency','yes'),'yes'):
		print "checking model consistency"
		if solutionstring.lower() == 'flaimsolution':
			md.private.isconsistent=True
			md.mesh.checkconsistency(md,solutionstring)
			md.flaim.checkconsistency(md,solutionstring)
			if not md.private.isconsistent:
				raise RuntimeError("Model not consistent, see messages above.")
		else:
			ismodelselfconsistent(md)

	#First, build a runtime name that is unique
	restart=options.getfieldvalue('restart','')
	if restart == 1:
		pass #do nothing
	else:
		if restart:
			md.private.runtimename=restart
		else:
			if options.getfieldvalue('runtimename',True):
				c=datetime.datetime.now()
				md.private.runtimename="%s-%02i-%02i-%04i-%02i-%02i-%02i-%i" % (md.miscellaneous.name,c.month,c.day,c.year,c.hour,c.minute,c.second,os.getpid())
			else:
				md.private.runtimename=md.miscellaneous.name 

	#if running qmu analysis, some preprocessing of dakota files using models
	#fields needs to be carried out. 
	if md.qmu.isdakota:
		md=preqmu(md,options)

	#flaim analysis
	if solutionstring.lower() == 'flaimSolution':
		solutionstring='FlaimSolution'
		md=flaim_sol(md,options)
		md.private.solution=solutionstring
		return md

	#Do we load results only?
	if options.getfieldvalue('loadonly',False):
		md=loadresultsfromcluster(md)
		return md


	#Write all input files
	marshall(md)                                           # bin file
	md.toolkits.ToolkitsFile(md.miscellaneous.name+'.toolkits')    # toolkits file
	cluster.BuildQueueScript(md.private.runtimename,md.miscellaneous.name,md.private.solution,md.settings.io_gather,md.debug.valgrind,md.debug.gprof,md.qmu.isdakota)    # queue file

	#Stop here if batch mode
	if m.strcmpi(options.getfieldvalue('batch','no'),'yes'):
		print 'batch mode requested: not launching job interactively'
		print 'launch solution sequence on remote cluster by hand'
		return md

	#Upload all required files: 
	modelname = md.miscellaneous.name
	filelist  = [modelname+'.bin ',modelname+'.toolkits ',modelname+'.queue ']
	if md.qmu.isdakota:
		filelist.append(modelname+'.qmu.in')

	if not restart:
		cluster.UploadQueueJob(md.miscellaneous.name,md.private.runtimename,filelist)
	
	#Launch job
	cluster.LaunchQueueJob(md.miscellaneous.name,md.private.runtimename,filelist,restart,batch)

	#wait on lock
	if md.settings.waitonlock>0:
		#we wait for the done file
		islock=waitonlock(md)
		if islock==0:    #no results to be loaded
			print 'The results must be loaded manually with md=loadresultsfromcluster(md).'
		else:            #load results
			print 'loading results from cluster'
			md=loadresultsfromcluster(md)

	#post processes qmu results if necessary
	if md.qmu.isdakota:
		if not strncmpi(options['keep'],'y',1):
			shutil.rmtree('qmu'+str(os.getpid()))

	return md
